"""
Defines base Redmine resource manager class and its infrastructure.
"""

from .. import resultsets, exceptions


class ResourceManager:
    """
    Manages given Redmine resource class with the help of redmine object.
    """
    def __init__(self, redmine, resource_class):
        """
        :param redmine.Redmine redmine: (required). Redmine object.
        :param resources.BaseResource resource_class: (required). Resource class.
        """
        self.url = ''
        self.params = {}
        self.container = None
        self.redmine = redmine
        self.resource_class = resource_class

    def to_resource(self, resource):
        """
        Converts resource data to Resource object.

        :param dict resource: (required). Resource data.
        """
        return self.resource_class(self, resource)

    def to_resource_set(self, resources):
        """
        Converts an iterable with resources data to ResourceSet object.

        :param resources: (required). Resource data.
        :type resources: list or tuple
        """
        return resultsets.ResourceSet(self, resources)

    def new(self):
        """
        Returns new empty Resource object.
        """
        return self.to_resource({})

    def new_manager(self, resource_name, **params):
        """
        Returns new ResourceManager object.

        :param string resource_name: (required). Resource name.
        :param dict params: (optional). Parameters used for resources retrieval.
        """
        manager = getattr(self.redmine, resource_name)
        manager.params = params
        return manager

    def _construct_get_url(self, path):
        """
        Constructs URL for get method.

        :param string path: absolute URL path.
        """
        return self.redmine.url + path

    def _prepare_get_request(self, request):
        """
        Makes the necessary preparations for get request data.

        :param dict request: Request data.
        """
        return self.resource_class.bulk_decode(request, self)

    def get(self, resource_id, **params):
        """
        Returns a Resource object from Redmine by resource id.

        :param resource_id: (required). Resource id.
        :type resource_id: int or string
        :param dict params: (optional). Parameters used for resource retrieval.
        """
        if self.resource_class.query_one is None or self.resource_class.container_one is None:
            operation = self.all if self.resource_class.query_all else self.filter
            resource = operation(**params).get(resource_id, None)

            if resource is None:
                raise exceptions.ResourceNotFoundError

            return resource

        try:
            self.url = self._construct_get_url(self.resource_class.query_one.format(resource_id, **params))
        except KeyError as e:
            raise exceptions.ValidationError(f'{e} argument is required')

        self.params = self._prepare_get_request(params)
        self.container = self.resource_class.container_one

        try:
            response = self.redmine.engine.request('get', self.url, params=self.params)
        except exceptions.ResourceNotFoundError as e:
            if self.resource_class.requirements:
                raise exceptions.ResourceRequirementsError(self.resource_class.requirements)
            raise e

        return self._process_get_response(self.params, response)

    def _process_get_response(self, request, response):
        """
        Processes get response and constructs resource object.

        :param dict request: Original request data.
        :param any response: Response received from Redmine for this request data.
        """
        return self.to_resource(response[self.container])

    def all(self, **params):
        """
        Returns a ResourceSet object with all Resource objects.

        :param dict params: (optional). Parameters used for resources retrieval.
        """
        if self.resource_class.query_all is None or self.resource_class.container_all is None:
            raise exceptions.ResourceBadMethodError

        self.url = self.redmine.url + self.resource_class.query_all
        self.params = self.resource_class.bulk_decode(params, self)
        self.container = self.resource_class.container_all
        return resultsets.ResourceSet(self)

    def filter(self, **filters):
        """
        Returns a ResourceSet object with Resource objects filtered by a dict of filters.

        :param dict filters: (optional). Filters used for resources retrieval.
        """
        if self.resource_class.query_filter is None or self.resource_class.container_filter is None:
            raise exceptions.ResourceBadMethodError

        if not filters:
            raise exceptions.ResourceNoFiltersProvidedError

        try:
            self.url = self.redmine.url + self.resource_class.query_filter.format(**filters)
            self.container = self.resource_class.container_filter.format(**filters)
        except KeyError:
            raise exceptions.ResourceFilterError

        self.params = self.resource_class.bulk_decode(filters, self)
        return resultsets.ResourceSet(self)

    def _construct_create_url(self, path):
        """
        Constructs URL for create method.

        :param string path: absolute URL path.
        """
        return self.redmine.url + path

    def _prepare_create_request(self, request):
        """
        Makes the necessary preparations for create request data.

        :param dict request: Request data.
        """
        return {self.container: self.resource_class.bulk_decode(request, self)}

    def create(self, **fields):
        """
        Creates a new resource in Redmine and returns created Resource object on success.

        :param dict fields: (optional). Fields used for resource creation.
        """
        if self.resource_class.query_create is None or self.resource_class.container_create is None:
            raise exceptions.ResourceBadMethodError

        if not fields:
            raise exceptions.ResourceNoFieldsProvidedError

        try:
            url = self._construct_create_url(self.resource_class.query_create.format(**fields))
        except KeyError as e:
            raise exceptions.ValidationError(f'{e} field is required')

        self.params = self.resource_class.query_create.formatter.used_kwargs
        self.container = self.resource_class.container_create
        request = self._prepare_create_request(self.resource_class.query_create.formatter.unused_kwargs)
        response = self.redmine.engine.request(self.resource_class.http_method_create, url, data=request)

        if response is None:
            return None

        resource = self._process_create_response(request, response)

        if self.resource_class.query_one is not None:
            self.url = self.redmine.url + self.resource_class.query_one.format(resource.internal_id, **fields)

        return resource

    def _process_create_response(self, request, response):
        """
        Processes create response and constructs resource object.

        :param dict request: Original request data.
        :param any response: Response received from Redmine for this request data.
        """
        return self.to_resource(response[self.container])

    def _construct_update_url(self, path):
        """
        Constructs URL for update method.

        :param string path: absolute URL path.
        """
        return self.redmine.url + path

    def _prepare_update_request(self, request):
        """
        Makes the necessary preparations for update request data.

        :param dict request: Request data.
        """
        return {self.container: self.resource_class.bulk_decode(request, self)}

    def update(self, resource_id, **fields):
        """
        Updates a Resource object by resource id.

        :param resource_id: (required). Resource id.
        :type resource_id: int or string
        :param dict fields: (optional). Fields that will be updated for the resource.
        """
        if self.resource_class.query_update is None or self.resource_class.container_update is None:
            raise exceptions.ResourceBadMethodError

        if not fields:
            raise exceptions.ResourceNoFieldsProvidedError

        try:
            query_update = self.resource_class.query_update.format(resource_id, **fields)
        except KeyError as e:
            param = e.args[0]

            if param in self.params:
                fields[param] = self.params[param]
                query_update = self.resource_class.query_update.format(resource_id, **fields)
            else:
                raise exceptions.ValidationError(f'{e} argument is required')

        self.params.update(self.resource_class.query_update.formatter.used_kwargs)
        self.container = self.resource_class.container_update
        url = self._construct_update_url(query_update)
        request = self._prepare_update_request(self.resource_class.query_update.formatter.unused_kwargs)
        response = self.redmine.engine.request(self.resource_class.http_method_update, url, data=request)

        if response is None:
            return None

        return self._process_update_response(request, response)

    def _process_update_response(self, request, response):
        """
        Processes update response.

        :param dict request: Original request data.
        :param any response: Response received from Redmine for this request data.
        """
        return response

    def _construct_delete_url(self, path):
        """
        Constructs URL for delete method.

        :param string path: absolute URL path.
        """
        return self.redmine.url + path

    def _prepare_delete_request(self, request):
        """
        Makes the necessary preparations for delete request data.

        :param dict request: Request data.
        """
        return self.resource_class.bulk_decode(request, self)

    def delete(self, resource_id, **params):
        """
        Deletes a Resource object by resource id.

        :param resource_id: (required). Resource id.
        :type resource_id: int or string
        :param dict params: (optional). Parameters used for resource deletion.
        """
        if self.resource_class.query_delete is None:
            raise exceptions.ResourceBadMethodError

        try:
            url = self._construct_delete_url(self.resource_class.query_delete.format(resource_id, **params))
        except KeyError as e:
            raise exceptions.ValidationError(f'{e} argument is required')

        request = self._prepare_delete_request(params)
        response = self.redmine.engine.request(self.resource_class.http_method_delete, url, params=request)

        if response is None:
            return None

        return self._process_delete_response(request, response)

    def _process_delete_response(self, request, response):
        """
        Processes delete response.

        :param dict request: Original request data.
        :param any response: Response received from Redmine for this request data.
        """
        return response

    def search(self, query, **options):
        """
        Searches for Resources using a query.

        :param string query: (required). What to search.
        :param dict options: (optional). Dictionary of search options.
        """
        if self.resource_class.search_hints is None:
            raise exceptions.ResourceBadMethodError

        container = self.resource_class.container_all or self.resource_class.container_filter
        results = self.redmine.search(query, **dict(resources=[container], **options))
        return results.get(container) if results is not None else results

    def __repr__(self):
        """
        Official representation of a ResourceManager object.
        """
        return f'<redminelib.managers.{self.__class__.__name__} object for {self.resource_class.__name__} resource>'



"""
Defines ResourceSet objects that can be used to represent a set of resources.
"""

import operator
import functools
import itertools

from . import lookups, exceptions


class BaseResourceSet:
    """
    Defines basic functionality for a ResourceSet object.
    """
    def __init__(self, manager, resources=None, limit=0, offset=0, total_count=None):
        """
        :param managers.ResourceManager manager: (required). ResourceManager object.
        :param resources: (optional). Iterable of resources.
        :type resources: list or tuple
        :param int limit: (optional). Resource limit.
        :param int offset: (optional). Resource offset.
        :param int total_count: (optional). How many resources are there available in Redmine.
        """
        self.manager = manager
        self.limit = limit
        self.offset = offset
        self._resources = resources
        self._total_count = total_count
        self._is_sliced = False

    @property
    def total_count(self):
        """
        Returns total count of available resources in Redmine, this is known only after ResourceSet evaluation.
        """
        if self._total_count is None:
            if self._resources is None:
                raise exceptions.ResultSetTotalCountError
            else:
                self._total_count = len(self)

        return self._total_count

    def export(self, fmt, savepath=None, filename=None, columns=None, encoding='UTF-8'):
        """
        Exports all resources from resource set to requested format if Resource supports that.

        :param string fmt: (required). Format to use for export, e.g. atom, csv, txt, pdf, html etc.
        :param string savepath: (optional). Path where to save the file.
        :param string filename: (optional). Name that will be used for the file.
        :param columns: (optional). Iterable of column names, "all_gui" for GUI behaviour or "all" for all columns.
        :param encoding: (optional). Encoding that will be used by Redmine for the result file.
        :type columns: iterable or string
        """
        if self.manager.resource_class.query_all_export is None:
            raise exceptions.ExportNotSupported

        url = self.manager.redmine.url + self.manager.resource_class.query_all_export.format(
                format=fmt, **self.manager.params)

        params = dict(self.manager.resource_class.query_all_export.formatter.unused_kwargs, encoding=encoding)

        if columns is not None:
            if columns == 'all':
                columns = ['all', 'all_inline'] + self.manager.resource_class.extra_export_columns

                if self.manager.redmine.ver is not None and self.manager.redmine.ver < (3, 4, 0):
                    params.update(dict.fromkeys(self.manager.resource_class.extra_export_columns, 1), columns='all')
            elif 'all_gui' in columns:
                if columns == 'all_gui':
                    columns = ['all', 'all_inline']

                    if self.manager.redmine.ver is not None and self.manager.redmine.ver < (3, 4, 0):
                        params['columns'] = 'all'
                else:
                    if self.manager.redmine.ver is not None and self.manager.redmine.ver < (3, 4, 0):
                        params.update(dict.fromkeys(columns, 1), columns='all')

                    columns = list(columns) + ['all', 'all_inline']

            # Redmine >= 3.4.0 happily accepts c[] array with column names
            # for older versions the above hack with params is being used
            params['c[]'] = columns

        try:
            return self.manager.redmine.download(url, savepath, filename, params=params)
        except exceptions.UnknownError as e:
            if e.status_code == 406:
                raise exceptions.ExportFormatNotSupportedError
            raise e

    def _resource_cls(self, cls, resources, **kwargs):
        """
        Returns a new resource set class instance defined by cls, filled with resources and loaded with kwargs.

        :param any cls: (required). Resource set class.
        :param resources: (required). Iterable of resources.
        :type resources: list or tuple
        :param dict kwargs: (optional). Additional keyword arguments if any.
        """
        return cls(self.manager, resources=resources, limit=self.limit, offset=self.offset,
                   total_count=self._total_count, **kwargs)

    def __getitem__(self, item):
        """
        Sets limit and offset or returns a Resource by requested index.
        """
        if isinstance(item, slice):
            self.limit = item.stop
            self.offset = item.start
            self._is_sliced = True
        elif isinstance(item, int):
            try:
                return next(itertools.islice(self, item, item + 1))
            except StopIteration:
                raise exceptions.ResourceSetIndexError

        if self._resources is not None and self._is_sliced:
            return self._resource_cls(self.__class__, [resource for resource in BaseResourceSet.__iter__(self)])

        return self

    def __iter__(self):
        """
        Returns requested resources in a lazy fashion.
        """
        # If this is the first time we are evaluating the ResourceSet
        # all the hard part will be done by the active Engine object
        if self._resources is None:
            self.manager.params.setdefault('limit', self.limit)
            self.manager.params.setdefault('offset', self.offset)

            try:
                self._resources, self._total_count = self.manager.redmine.engine.bulk_request(
                    'get', self.manager.url, self.manager.container, **self.manager.params)
            except exceptions.ResourceNotFoundError as e:
                if self.manager.resource_class.requirements:
                    raise exceptions.ResourceRequirementsError(self.manager.resource_class.requirements)
                raise e

            resources = self._resources
        # Otherwise ResourceSet object should handle slicing by itself
        elif self._is_sliced:
            offset = self.offset or None

            if not self.limit:
                limit = None
            elif self.limit and not self.offset:
                limit = self.limit
            else:
                limit = self.limit + self.offset

            resources = self._resources[offset:limit]
        else:
            resources = self._resources

        self._is_sliced = False
        return (resource for resource in resources)

    def __len__(self):
        """
        Allows len() to be called on a ResourceSet object.
        """
        return sum(1 for _ in self)

    def __repr__(self):
        """
        Official representation of a ResourceSet object.
        """
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__} object '
            f'with {self.manager.resource_class.__name__} resources>'
        )


class ResourceSet(BaseResourceSet):
    """
    Represents a set of Redmine resources as objects.
    """
    def get(self, resource_id, default=None):
        """
        Returns a single Resource from a ResourceSet by resource id.

        :param resource_id: (required). Resource id.
        :type resource_id: int or string
        :param none default: (optional). What to return if Resource wasn't found.
        """
        for resource in super().__iter__():
            if resource_id == resource[self.manager.resource_class.internal_id_key]:
                return self.manager.to_resource(resource)

        return default

    def filter(self, **filters):
        """
        Returns a new filtered ResourceSet with requested filters applied.

        :param dict filters: (required). Filters used for resources retrieval.
        """
        if not filters:
            raise exceptions.ResourceNoFiltersProvidedError

        reducers = []

        for f in filters:
            fields = f.split('__')
            lookup = fields[-1]

            reducer = {
                'fields': fields,
                'value': filters[f],
                'lookup': lookups.registry['exact'],
                'lookup_name': lookup,
                'filter_name': f,
            }

            if lookup in lookups.registry:
                reducer['fields'] = fields[:-1]
                reducer['lookup'] = lookups.registry[lookup]

            reducers.append(reducer)

        resources = []

        for resource in super().__iter__():
            for r in reducers:
                try:
                    if not r['lookup'](functools.reduce(operator.getitem, r['fields'], resource), r['value']):
                        break
                except KeyError:
                    break
                except TypeError:
                    raise exceptions.ResourceSetFilterLookupError(r['lookup_name'], r['filter_name'])
            else:
                resources.append(resource)

        return self._resource_cls(ResourceSet, resources)

    def update(self, **fields):
        """
        Updates fields of all resources in a ResourceSet with the given values.

        :param dict fields: (optional). Fields in resources that will be updated.
        """
        resources = []

        for resource in self:
            for field in fields:
                setattr(resource, field, fields[field])

            resources.append(resource.save().raw())

        return self._resource_cls(ResourceSet, resources)

    def delete(self):
        """
        Deletes all resources in a ResourceSet.
        """
        for resource in self:
            resource.delete()

        self._resources = None
        return True

    def values(self, *fields):
        """
        Returns ResourceSet as an iterable of dictionaries.

        :param fields: (optional). Iterable which sets field names each resource will contain.
        :type fields: list or tuple
        """
        if fields:
            for resource in super().__iter__():
                yield {field: resource[field] for field in fields if field in resource}
        else:
            for resource in super().__iter__():
                yield resource

    def values_list(self, *fields, **kwargs):
        """
        Returns ResourceSet as an iterable of tuples with Resource values or single values if flattened.

        :param fields: (optional). Iterable which sets field names each resource will contain.
        :type fields: list or tuple
        :param dict kwargs: (optional). If fields contain single field, setting flat=True will flatten the result.
        """
        flat = kwargs.pop('flat', False)

        if fields:
            if flat and len(fields) == 1:
                for resource in super().__iter__():
                    yield resource.get(fields[0])
            else:
                for resource in super().__iter__():
                    yield tuple(resource[field] for field in fields if field in resource)
        else:
            for resource in super().__iter__():
                yield tuple(resource.values())

    def __iter__(self):
        """
        Returns requested resources in a lazy fashion.
        """
        return (self.manager.to_resource(resource) for resource in super().__iter__())



"""
Defines standard Redmine resources managers.
"""

from . import ResourceManager
from .. import exceptions


class ProjectManager(ResourceManager):
    def __getattr__(self, attr):
        if attr in ('close', 'reopen', 'archive', 'unarchive'):
            if self.redmine.ver is not None and self.redmine.ver < (5, 0, 0):
                raise exceptions.VersionMismatchError(f'Project {attr}')

            return lambda resource_id: self.redmine.engine.request(
                'put', f'{self.redmine.url}{self.resource_class.query_one.format(resource_id)[:-5]}/{attr}.json')

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")


class IssueManager(ResourceManager):
    def copy(self, issue_id, link_original=True, include=(), **fields):
        fields['_copy'] = {'copy_from': issue_id}

        if link_original:
            fields['_copy']['link_copy'] = '1'

        if include is not None:
            for i in include or ('subtasks', 'attachments'):
                fields['_copy'][f'copy_{i}'] = '1'

        return self.create(**fields)

    def _prepare_create_request(self, request):
        request = super()._prepare_create_request(request)
        request.update(request[self.container].pop('_copy', {}))
        return request


class FileManager(ResourceManager):
    def _process_create_response(self, request, response):
        if response is True:
            response = {self.container: {'id': int(request[self.container]['token'].split('.')[0])}}

        return super()._process_create_response(request, response)


class WikiPageManager(ResourceManager):
    def _process_create_response(self, request, response):
        if response is True:
            raise exceptions.ValidationError('Resource already exists')  # issue #182

        return super()._process_create_response(request, response)


class UserManager(ResourceManager):
    @staticmethod
    def _check_custom_url(path):
        if path.endswith('/me.json'):
            path = '/my/account.json'

        return path

    def _construct_get_url(self, path):
        return super()._construct_get_url(self._check_custom_url(path))

    def all(self, **params):
        resourceset = super().all(**params)

        if self.redmine.ver is not None:  # https://www.redmine.org/issues/32090#note-6
            if self.redmine.ver >= (5, 1, 2):
                resourceset.manager.url = f'{resourceset.manager.url}*'
            elif self.redmine.ver in ((5, 1, 0), (5, 1, 1)):
                resourceset.manager.url = (f'{resourceset.manager.url[:-7]}f[]=status_id&'
                                           f'op[status_id]==&v[status_id][]=1&v[status_id][]=2&v[status_id][]=3')

        return resourceset

    def _prepare_create_request(self, request):
        request = super()._prepare_create_request(request)
        request['send_information'] = request[self.container].pop('send_information', False)
        return request

    def _construct_update_url(self, path):
        return super()._construct_update_url(self._check_custom_url(path))

    def _prepare_update_request(self, request):
        request = super()._prepare_update_request(request)
        request['send_information'] = request[self.resource_class.container_update].pop('send_information', False)
        return request


class NewsManager(ResourceManager):
    def _process_create_response(self, request, response):
        if response is True:
            response = {self.container: self.redmine.news.filter(**self.params)[0].raw()}

        return super()._process_create_response(request, response)



"""
Python-Redmine tries its best to provide human-readable errors in all situations.
This is a list of all exceptions or warnings that Python-Redmine can throw/raise.
"""


class BaseRedmineWarning(Warning):
    """
    Base warning class for Redmine warnings.
    """


class PerformanceWarning(BaseRedmineWarning):
    """
    Warning raised when there's a possible performance impact.
    """


class BaseRedmineError(Exception):
    """
    Base exception class for Redmine exceptions.
    """


class ResourceError(BaseRedmineError):
    """
    Unsupported Redmine resource exception.
    """
    def __init__(self):
        super().__init__('Unsupported Redmine resource')


class NoFileError(BaseRedmineError):
    """
    File doesn't exist or is empty exception.
    """
    def __init__(self):
        super().__init__("Can't upload a file that doesn't exist or is empty")


class FileObjectError(BaseRedmineError):
    """
    File-like object isn't supported as it doesn't support the read(size) method.
    """
    def __init__(self):
        super().__init__("File-like object doesn't support the read(size) method")


class ResourceNotFoundError(BaseRedmineError):
    """
    Requested resource doesn't exist.
    """
    def __init__(self):
        super().__init__("Requested resource doesn't exist")


class ConflictError(BaseRedmineError):
    """
    Resource version on the server is newer than on the client.
    """
    def __init__(self):
        super().__init__('Resource version on the server is newer than on the client')


class AuthError(BaseRedmineError):
    """
    Invalid authentication details.
    """
    def __init__(self):
        super().__init__('Invalid authentication details')


class ImpersonateError(BaseRedmineError):
    """
    Invalid impersonate login provided.
    """
    def __init__(self):
        super().__init__("Impersonate login provided doesn't exist or isn't active")


class ServerError(BaseRedmineError):
    """
    Redmine internal error.
    """
    def __init__(self):
        super().__init__('Redmine returned internal error, check Redmine logs for details')


class RequestEntityTooLargeError(BaseRedmineError):
    """
    Size of the request exceeds the capacity limit on the server.
    """
    def __init__(self):
        super().__init__(
            "The requested resource doesn't allow POST requests or the size of the request exceeds the capacity limit")


class UnknownError(BaseRedmineError):
    """
    Redmine returned unknown error.
    """
    def __init__(self, status_code):
        self.status_code = status_code
        super().__init__(f'Redmine returned unknown error with the status code {status_code}')


class ValidationError(BaseRedmineError):
    """
    Redmine validation errors occurred on create/update resource.
    """
    def __init__(self, error):
        super().__init__(error)


class ResourceSetIndexError(BaseRedmineError):
    """
    Index doesn't exist in the ResourceSet.
    """
    def __init__(self):
        super().__init__('Resource not available by requested index')


class ResourceSetFilterLookupError(BaseRedmineError):
    """
    Resource set filter method received an invalid lookup in one of the filters.
    """
    def __init__(self, lookup, f):
        super().__init__(f'Received an invalid lookup "{lookup}" in "{f}" filter')


class ResourceBadMethodError(BaseRedmineError):
    """
    Resource doesn't support the requested method.
    """
    def __init__(self):
        super().__init__("Resource doesn't support the requested method")


class ResourceFilterError(BaseRedmineError):
    """
    Resource doesn't support requested filter(s).
    """
    def __init__(self):
        super().__init__("Resource doesn't support requested filter(s)")


class ResourceNoFiltersProvidedError(BaseRedmineError):
    """
    No filter(s) provided.
    """
    def __init__(self):
        super().__init__('Resource needs some filters to be filtered on')


class ResourceNoFieldsProvidedError(BaseRedmineError):
    """
    No field(s) provided.
    """
    def __init__(self):
        super().__init__('Resource needs some fields to be set to be created/updated')


class ResourceAttrError(BaseRedmineError, AttributeError):
    """
    Resource doesn't have the requested attribute.
    """
    def __init__(self):
        super().__init__("Resource doesn't have the requested attribute")


class ReadonlyAttrError(BaseRedmineError):
    """
    Resource can't set attribute that is read only.
    """
    def __init__(self):
        super().__init__("Can't set read only attribute")


class VersionFormatError(BaseRedmineError):
    """
    Version format provided isn't supported. SemVer is the only format accepted.
    """
    def __init__(self, version):
        super().__init__(
            f"Version in the {version} format isn't supported, please provide numeric version in the form of X.X.X")


class VersionMismatchError(BaseRedmineError):
    """
    Feature isn't supported on specified Redmine version.
    """
    def __init__(self, feature):
        super().__init__(f"{feature} isn't supported on specified Redmine version")


class ResourceVersionMismatchError(VersionMismatchError):
    """
    Resource isn't supported on specified Redmine version.
    """
    def __init__(self):
        super().__init__('Resource')


class ResultSetTotalCountError(BaseRedmineError):
    """
    ResultSet hasn't been yet evaluated and cannot yield a total_count.
    """
    def __init__(self):
        super().__init__('Total count is unknown before evaluation')


class CustomFieldValueError(BaseRedmineError):
    """
    Custom fields should be passed as a list of dictionaries.
    """
    def __init__(self):
        super().__init__(
            "Custom fields should be passed as a list of dictionaries in the form of [{'id': 1, 'value': 'foo'}]")


class ResourceRequirementsError(BaseRedmineError):
    """
    Resource requires specific Redmine plugin(s) to function.
    """
    def __init__(self, requirements):
        reqs = []

        for req in requirements:
            if isinstance(req, (list, tuple)):
                reqs.append(' >= '.join([req[0], '.'.join(map(str, req[1]))]))
            else:
                reqs.append(req)

        super().__init__(f"The following requirements must be installed for resource to function: {', '.join(reqs)}")


class FileUrlError(BaseRedmineError):
    """
    URL provided to download a file can't be parsed.
    """
    def __init__(self):
        super().__init__("URL provided to download a file can't be parsed")


class ForbiddenError(BaseRedmineError):
    """
    Requested resource is forbidden.
    """
    def __init__(self):
        super().__init__('Requested resource is forbidden')


class JSONDecodeError(BaseRedmineError):
    """
    Unable to decode received JSON.
    """
    def __init__(self, response):
        self.response = response
        super().__init__(
            'Unable to decode received JSON, you can inspect exception\'s '
            '"response" attribute to find out what the response was')


class ExportNotSupported(BaseRedmineError):
    """
    Export functionality not supported by resource.
    """
    def __init__(self):
        super().__init__('Export functionality not supported by resource')


class ExportFormatNotSupportedError(BaseRedmineError):
    """
    The given format isn't supported by resource.
    """
    def __init__(self):
        super().__init__("The given format isn't supported by resource")


class HTTPProtocolError(BaseRedmineError):
    """
    Wrong HTTP protocol usage.
    """
    def __init__(self):
        super().__init__('Protocol redirect detected, Redmine URL expects HTTPS, but code uses HTTP or vice versa')


class TimezoneError(BaseRedmineError):
    """
    Timezone is neither a string, suitable for a strptime %z, nor is an instance of tzinfo class.
    """
    def __init__(self):
        super().__init__(
            'Timezone has to be either a ±HHMM string, e.g. -0800 or +0545, or an instance of datetime.tzinfo class')


class EngineClassError(BaseRedmineError):
    """
    Engine isn't a class or isn't a BaseEngine subclass.
    """
    def __init__(self):
        super().__init__("Engine isn't a class or isn't a BaseEngine subclass")



"""
Defines standard Redmine resources and resource mappings.
"""

from . import BaseResource
from .. import managers, exceptions


class Project(BaseResource):
    redmine_version = (1, 0, 0)
    container_all = 'projects'
    container_one = 'project'
    container_create = 'project'
    container_update = 'project'
    query_all_export = '/projects.{format}'
    query_all = '/projects.json'
    query_one = '/projects/{}.json'
    query_create = '/projects.json'
    query_update = '/projects/{}.json'
    query_delete = '/projects/{}.json'
    search_hints = ['project']
    manager_class = managers.ProjectManager

    _repr = [['id', 'name'], ['title']]
    _includes = ['trackers', 'issue_categories', 'enabled_modules', 'time_entry_activities', 'issue_custom_fields']
    _relations = ['wiki_pages', 'memberships', 'issue_categories', 'time_entries', 'versions',
                  'news', 'issues', 'files']
    _unconvertible = BaseResource._unconvertible + ['identifier', 'status']
    _update_readonly = BaseResource._update_readonly + ['identifier']
    _resource_map = {'default_version': 'Version', 'default_assignee': 'User'}
    _resource_set_map = {
        'custom_fields': 'CustomField',
        'trackers': 'Tracker',
        'issue_categories': 'IssueCategory',
        'wiki_pages': 'WikiPage',
        'memberships': 'ProjectMembership',
        'time_entries': 'TimeEntry',
        'versions': 'Version',
        'news': 'News',
        'issues': 'Issue',
        'files': 'File',
        'time_entry_activities': 'Enumeration',
        'issue_custom_fields': 'CustomField',
    }
    _single_attr_id_map = {'parent_id': 'parent'}
    _multiple_attr_id_map = {'tracker_ids': 'trackers'}

    @property
    def url(self):
        return self.manager.redmine.url + self.query_one.format(self.identifier)[:-5]

    @classmethod
    def encode(cls, attr, value, manager):
        if attr == 'enabled_modules':
            return attr, [module['name'] for module in value]

        return super().encode(attr, value, manager)

    def __getattr__(self, attr):
        if attr in ('close', 'reopen', 'archive', 'unarchive'):
            return lambda: getattr(self.manager, attr)(self.internal_id)

        return super().__getattr__(attr)


class Issue(BaseResource):
    redmine_version = (1, 0, 0)
    container_all = 'issues'
    container_one = 'issue'
    container_filter = 'issues'
    container_create = 'issue'
    container_update = 'issue'
    query_all_export = '/issues.{format}'
    query_one_export = '/issues/{}.{format}'
    query_all = '/issues.json?status_id=*'
    query_one = '/issues/{}.json'
    query_filter = '/issues.json'
    query_create = '/projects/{project_id}/issues.json'
    query_update = '/issues/{}.json'
    query_delete = '/issues/{}.json'
    search_hints = ['issue', 'issue closed', 'issue-closed']
    extra_export_columns = ['description', 'last_notes']
    manager_class = managers.IssueManager

    _repr = [['id', 'subject'], ['title'], ['id']]
    _includes = ['children', 'attachments', 'relations', 'changesets', 'journals', 'watchers', 'allowed_statuses']
    _relations = ['relations', 'time_entries']
    _unconvertible = BaseResource._unconvertible + ['subject', 'notes']
    _create_readonly = BaseResource._create_readonly + ['spent_hours']
    _update_readonly = _create_readonly[:]
    _resource_map = {
        'project': 'Project',
        'tracker': 'Tracker',
        'status': 'IssueStatus',
        'priority': 'Enumeration',
        'author': 'User',
        'assigned_to': 'User',
        'category': 'IssueCategory',
        'fixed_version': 'Version',
    }
    _resource_set_map = {
        'custom_fields': 'CustomField',
        'attachments': 'Attachment',
        'journals': 'IssueJournal',
        'children': 'Issue',
        'relations': 'IssueRelation',
        'watchers': 'User',
        'time_entries': 'TimeEntry',
        'allowed_statuses': 'IssueStatus',
    }
    _single_attr_id_map = {
        'project_id': 'project',
        'tracker_id': 'tracker',
        'status_id': 'status',
        'priority_id': 'priority',
        'category_id': 'category',
        'fixed_version_id': 'fixed_version',
        'assigned_to_id': 'assigned_to',
        'parent_issue_id': 'parent',
    }
    _multiple_attr_id_map = {'watcher_user_ids': 'watchers'}

    class Watcher:
        """
        An issue watcher implementation.
        """
        def __init__(self, issue):
            self._redmine = issue.manager.redmine
            self._issue_id = issue.internal_id

            if self._redmine.ver is not None and self._redmine.ver < (2, 3, 0):
                raise exceptions.ResourceVersionMismatchError

        def add(self, user_id):
            """
            Adds user to issue watchers list.

            :param int user_id: (required). User id.
            """
            url = f'{self._redmine.url}/issues/{self._issue_id}/watchers.json'
            return self._redmine.engine.request('post', url, data={'user_id': user_id})

        def remove(self, user_id):
            """
            Removes user from issue watchers list.

            :param int user_id: (required). User id.
            """
            url = f'{self._redmine.url}/issues/{self._issue_id}/watchers/{user_id}.json'
            return self._redmine.engine.request('delete', url)

    def __getattr__(self, attr):
        if attr == 'watcher':
            return Issue.Watcher(self)

        if attr == 'version':
            attr = 'fixed_version'

        return super().__getattr__(attr)

    def __setattr__(self, attr, value):
        if attr == 'version_id':
            attr = 'fixed_version_id'

        super().__setattr__(attr, value)

    @classmethod
    def decode(cls, attr, value, manager):
        if attr == 'version_id':
            return 'fixed_version_id', value
        elif attr == 'assigned_to_id' and value in (None, 0):
            return attr, ''
        elif attr == 'checklists':
            return 'checklists_attributes', value

        return super().decode(attr, value, manager)

    def copy(self, link_original=True, include=(), **fields):
        if 'project_id' not in fields and not self.is_new():
            fields['project_id'] = self._decoded_attrs['project']['id']

        return self.manager.copy(self.internal_id, link_original=link_original, include=include, **fields)


class TimeEntry(BaseResource):
    redmine_version = (1, 1, 0)
    container_all = 'time_entries'
    container_one = 'time_entry'
    container_filter = 'time_entries'
    container_create = 'time_entry'
    container_update = 'time_entry'
    query_all_export = '/time_entries.{format}'
    query_all = '/time_entries.json'
    query_one = '/time_entries/{}.json'
    query_filter = '/time_entries.json'
    query_create = '/time_entries.json'
    query_update = '/time_entries/{}.json'
    query_delete = '/time_entries/{}.json'

    _repr = [['id']]
    _resource_map = {'project': 'Project', 'issue': 'Issue', 'user': 'User', 'activity': 'Enumeration'}
    _resource_set_map = {'custom_fields': 'CustomField'}
    _single_attr_id_map = {'project_id': 'project', 'issue_id': 'issue', 'activity_id': 'activity'}

    @classmethod
    def decode(cls, attr, value, manager):
        if attr == 'from_date':
            attr = 'from'
        elif attr == 'to_date':
            attr = 'to'

        return super().decode(attr, value, manager)


class Enumeration(BaseResource):
    redmine_version = (2, 2, 0)
    container_filter = '{resource}'
    query_filter = '/enumerations/{resource}.json'
    query_url = '/enumerations/{}/edit'

    _resource_set_map = {'custom_fields': 'CustomField'}


class Attachment(BaseResource):
    redmine_version = (1, 3, 0)
    container_one = 'attachment'
    container_update = 'attachment'
    query_one = '/attachments/{}.json'
    query_update = '/attachments/{}.json'
    query_delete = '/attachments/{}.json'
    http_method_update = 'patch'

    _repr = [['id', 'filename'], ['id']]
    _resource_map = {'author': 'User'}

    def download(self, savepath=None, filename=None):
        return self.manager.redmine.download(self.content_url, savepath, filename)


class File(Attachment):
    redmine_version = (3, 4, 0)
    container_filter = 'files'
    container_create = 'file'
    query_filter = '/projects/{project_id}/files.json'
    query_create = '/projects/{project_id}/files.json'
    manager_class = managers.FileManager

    _resource_map = {'author': 'User', 'version': 'Version'}

    @classmethod
    def decode(cls, attr, value, manager):
        if attr == 'path':
            return 'token', manager.redmine.upload(value)['token']

        return super().decode(attr, value, manager)


class IssueJournal(BaseResource):
    redmine_version = (1, 0, 0)
    container_update = 'journal'
    query_update = '/journals/{}.json'

    _repr = [['id']]
    _unconvertible = ['notes']
    _resource_map = {'user': 'User'}


class WikiPage(BaseResource):
    internal_id_key = 'title'
    redmine_version = (2, 2, 0)
    container_filter = 'wiki_pages'
    container_one = 'wiki_page'
    container_create = 'wiki_page'
    container_update = 'wiki_page'
    query_one_export = '/projects/{project_id}/wiki/{}.{format}'
    query_filter = '/projects/{project_id}/wiki/index.json'
    query_one = '/projects/{project_id}/wiki/{}.json'
    query_create = '/projects/{project_id}/wiki/{title}.json'
    query_update = '/projects/{project_id}/wiki/{}.json'
    query_delete = '/projects/{project_id}/wiki/{}.json'
    search_hints = ['wiki-page']
    http_method_create = 'put'
    manager_class = managers.WikiPageManager

    _repr = [['title']]
    _includes = ['attachments']
    _unconvertible = BaseResource._unconvertible + ['title', 'text']
    _create_readonly = BaseResource._create_readonly + ['version']
    _update_readonly = _create_readonly[:]
    _resource_map = {'author': 'User'}
    _resource_set_map = {'attachments': 'Attachment'}
    _single_attr_id_map = {'project_id': 'project'}

    @classmethod
    def encode(cls, attr, value, manager):
        if attr == 'parent':
            value = manager.new_manager(cls.__name__, project_id=manager.params.get('project_id', 0)).to_resource(value)
            return attr, value

        return super().encode(attr, value, manager)

    def refresh(self, **params):
        return super().refresh(**dict(params, project_id=self.project_id))

    def post_update(self):
        self._encoded_attrs['version'] = self._decoded_attrs['version'] = self._decoded_attrs.get('version', 0) + 1

    def delete(self, **params):
        return super().delete(**dict(params, project_id=self.project_id))

    def export_url(self, fmt):
        return self.manager.redmine.url + self.query_one_export.format(
            self.internal_id, project_id=self.project_id, format=fmt)

    @property
    def project_id(self):
        return self.manager.params.get('project_id', 0)

    @property
    def url(self):
        return self.manager.redmine.url + self.query_one.format(self.internal_id, project_id=self.project_id)[:-5]

    def __getattr__(self, attr):
        # If a text attribute of a resource is missing, we should
        # refresh a resource automatically for user's convenience
        if attr == 'text' and attr not in self._decoded_attrs:
            self._decoded_attrs[attr] = self.refresh(itself=False).raw()[attr]

        return super().__getattr__(attr)

    def __int__(self):
        return self.version


class ProjectMembership(BaseResource):
    redmine_version = (1, 4, 0)
    container_filter = 'memberships'
    container_one = 'membership'
    container_update = 'membership'
    container_create = 'membership'
    query_filter = '/projects/{project_id}/memberships.json'
    query_one = '/memberships/{}.json'
    query_create = '/projects/{project_id}/memberships.json'
    query_update = '/memberships/{}.json'
    query_delete = '/memberships/{}.json'

    _repr = [['id']]
    _create_readonly = BaseResource._create_readonly + ['user', 'roles']
    _update_readonly = _create_readonly[:]
    _resource_map = {'project': 'Project', 'user': 'User', 'group': 'Group'}
    _resource_set_map = {'roles': 'Role'}
    _single_attr_id_map = {'project_id': 'project', 'user_id': 'user'}
    _multiple_attr_id_map = {'role_ids': 'roles'}


class IssueCategory(BaseResource):
    redmine_version = (1, 3, 0)
    container_filter = 'issue_categories'
    container_one = 'issue_category'
    container_update = 'issue_category'
    container_create = 'issue_category'
    query_filter = '/projects/{project_id}/issue_categories.json'
    query_one = '/issue_categories/{}.json'
    query_create = '/projects/{project_id}/issue_categories.json'
    query_update = '/issue_categories/{}.json'
    query_delete = '/issue_categories/{}.json'

    _resource_map = {'project': 'Project', 'assigned_to': 'User'}
    _single_attr_id_map = {'project_id': 'project', 'assigned_to_id': 'assigned_to'}


class IssueRelation(BaseResource):
    redmine_version = (1, 3, 0)
    container_filter = 'relations'
    container_one = 'relation'
    container_create = 'relation'
    query_filter = '/issues/{issue_id}/relations.json'
    query_one = '/relations/{}.json'
    query_create = '/issues/{issue_id}/relations.json'
    query_delete = '/relations/{}.json'

    _repr = [['id']]
    _single_attr_id_map = {'issue_id': 'issue'}


class Version(BaseResource):
    redmine_version = (1, 3, 0)
    container_filter = 'versions'
    container_one = 'version'
    container_create = 'version'
    container_update = 'version'
    query_filter = '/projects/{project_id}/versions.json'
    query_one = '/versions/{}.json'
    query_create = '/projects/{project_id}/versions.json'
    query_update = '/versions/{}.json'
    query_delete = '/versions/{}.json'

    _unconvertible = ['status']
    _resource_map = {'project': 'Project'}
    _resource_set_map = {'custom_fields': 'CustomField'}
    _single_attr_id_map = {'project_id': 'project'}


class User(BaseResource):
    redmine_version = (1, 1, 0)
    container_all = 'users'
    container_one = 'user'
    container_filter = 'users'
    container_create = 'user'
    container_update = 'user'
    query_all_export = '/users.{format}'
    query_all = '/users.json?status='
    query_one = '/users/{}.json'
    query_filter = '/users.json'
    query_create = '/users.json'
    query_update = '/users/{}.json'
    query_delete = '/users/{}.json'
    manager_class = managers.UserManager

    _repr = [['id', 'firstname', 'lastname'], ['id', 'name']]
    _includes = ['memberships', 'groups']
    _relations = ['issues', 'issues_assigned', 'issues_authored', 'time_entries']
    _relations_name = 'assigned_to'
    _unconvertible = ['status']
    _create_readonly = BaseResource._create_readonly + ['api_key', 'last_login_on']
    _update_readonly = _create_readonly[:]
    _resource_set_map = {
        'custom_fields': 'CustomField',
        'groups': 'Group',
        'memberships': 'ProjectMembership',
        'issues': 'Issue',
        'issues_assigned': 'Issue',
        'issues_authored': 'Issue',
        'time_entries': 'TimeEntry',
    }

    def __getattr__(self, attr):
        if attr in self._relations and attr not in self._encoded_attrs:
            if attr == 'issues_authored':
                self._relations_name = 'author'
            elif attr == 'time_entries':
                self._relations_name = 'user'

            value = super().__getattr__(attr)
            self._relations_name = 'assigned_to'
            return value

        return super().__getattr__(attr)


class Group(BaseResource):
    redmine_version = (2, 1, 0)
    container_all = 'groups'
    container_one = 'group'
    container_create = 'group'
    container_update = 'group'
    query_all = '/groups.json'
    query_one = '/groups/{}.json'
    query_create = '/groups.json'
    query_update = '/groups/{}.json'
    query_delete = '/groups/{}.json'

    _includes = ['memberships', 'users']
    _resource_set_map = {'memberships': 'ProjectMembership', 'users': 'User', 'custom_fields': 'CustomField'}
    _multiple_attr_id_map = {'user_ids': 'users'}

    class User:
        """
        A group user implementation.
        """
        def __init__(self, group):
            self._redmine = group.manager.redmine
            self._group_id = group.internal_id

        def add(self, user_id):
            """
            Adds user to a group.

            :param int user_id: (required). User id.
            """
            url = f'{self._redmine.url}/groups/{self._group_id}/users.json'
            return self._redmine.engine.request('post', url, data={'user_id': user_id})

        def remove(self, user_id):
            """
            Removes user from a group.

            :param int user_id: (required). User id.
            """
            url = f'{self._redmine.url}/groups/{self._group_id}/users/{user_id}.json'
            return self._redmine.engine.request('delete', url)

    def __getattr__(self, attr):
        if attr == 'user':
            return Group.User(self)

        return super().__getattr__(attr)


class Role(BaseResource):
    redmine_version = (1, 4, 0)
    container_all = 'roles'
    container_one = 'role'
    query_all = '/roles.json'
    query_one = '/roles/{}.json'


class News(BaseResource):
    redmine_version = (1, 1, 0)
    container_all = 'news'
    container_one = 'news'
    container_filter = 'news'
    container_create = 'news'
    container_update = 'news'
    query_all_export = '/news.{format}'
    query_all = '/news.json'
    query_one = '/news/{}.json'
    query_filter = '/news.json'
    query_create = '/projects/{project_id}/news.json'
    query_update = '/news/{}.json'
    query_delete = '/news/{}.json'
    search_hints = ['news']
    manager_class = managers.NewsManager

    _repr = [['id', 'title']]
    _includes = ['attachments', 'comments']
    _resource_map = {'project': 'Project', 'author': 'User'}
    _resource_set_map = {'attachments': 'Attachment'}
    _single_attr_id_map = {'project_id': 'project'}


class IssueStatus(BaseResource):
    redmine_version = (1, 3, 0)
    container_all = 'issue_statuses'
    query_all = '/issue_statuses.json'
    query_url = '/issue_statuses/{}/edit'

    _relations = ['issues']
    _relations_name = 'status'
    _resource_set_map = {'issues': 'Issue'}


class Tracker(BaseResource):
    redmine_version = (1, 3, 0)
    container_all = 'trackers'
    query_all = '/trackers.json'
    query_url = '/trackers/{}/edit'

    _relations = ['issues']
    _resource_set_map = {'issues': 'Issue'}


class Query(BaseResource):
    redmine_version = (1, 3, 0)
    container_all = 'queries'
    query_all = '/queries.json'
    query_url = '/projects/{}/issues?query_id={}'

    @property
    def url(self):
        return self.manager.redmine.url + self.query_url.format(
            self._decoded_attrs.get('project_id', 0), self.internal_id)


class CustomField(BaseResource):
    redmine_version = (2, 4, 0)
    container_all = 'custom_fields'
    query_all = '/custom_fields.json'
    query_url = '/custom_fields/{}/edit'

    _resource_set_map = {'trackers': 'Tracker', 'roles': 'Role'}

    def __getattr__(self, attr):
        # If custom field was created after the creation of the resource,
        # i.e. project, and it's not used in the resource, there will be
        # no value attribute defined, that is why we need to return '' or
        # we'll get an exception
        if attr == 'value' and attr not in self._decoded_attrs:
            return ''

        return super().__getattr__(attr)

    @classmethod
    def encode(cls, attr, value, manager):
        # Redmine <2.5.2 returns only single tracker instead of a list of
        # all available trackers, see http://www.redmine.org/issues/16739
        # for details
        if attr == 'trackers' and 'tracker' in value:
            value = [value['tracker']]

        return super().encode(attr, value, manager)



"""
Provides helper utilities.
"""

import copy
import string
import urllib.parse

from . import exceptions


def versiontuple(version):
    """
    Converts numeric SemVer version string to tuple.

    :param string version: (required). Version string to convert.
    """
    parts = str(version).split('.')

    if len(parts) != 3 or not all(part.isnumeric() for part in parts):
        raise exceptions.VersionFormatError(version)

    return tuple(int(part) for part in parts)


def merge_dicts(a, b):
    """
    Merges dicts a and b recursively into a new dict.

    :param dict a: (required).
    :param dict b: (required).
    """
    result = copy.deepcopy(a)

    for key, value in b.items():
        if isinstance(value, dict):
            result[key] = merge_dicts(value, a.get(key, {}))
        else:
            result[key] = value

    return result


class ResourceQueryFormatter(string.Formatter):
    """
    Quotes query and memorizes all arguments, used during string formatting.
    """
    def __init__(self):
        self.used_kwargs = {}
        self.unused_kwargs = {}

    def check_unused_args(self, used_args, args, kwargs):
        for item in used_args:
            if item in kwargs:
                self.used_kwargs[item] = kwargs.pop(item)

        self.unused_kwargs = kwargs

    def format_field(self, value, format_spec):
        return urllib.parse.quote(super().format_field(value, format_spec).encode('utf-8'))


class ResourceQueryStr(str):
    """
    Extends default string with additional formatting capabilities.
    """
    formatter = ResourceQueryFormatter()

    def format(self, *args, **kwargs):
        return self.formatter.format(self, *args, **kwargs)



from . import mock, BaseRedmineTestCase

from redminelib import exceptions

response = {
    'issues': [
        {'subject': 'Foo', 'id': 1, 'tracker_id': 1},
        {'subject': 'Bar', 'id': 2, 'tracker_id': 2},
        {'subject': 'Baz', 'id': 3, 'tracker_id': 3},
    ]
}


class ResultSetTestCase(BaseRedmineTestCase):
    def setUp(self):
        super().setUp()
        self.response.json = mock.Mock(return_value=response)

    def test_has_custom_repr(self):
        self.assertEqual(repr(self.redmine.issue.all()),
                         '<redminelib.resultsets.ResourceSet object with Issue resources>')

    def test_offset_limit_all(self):
        self.response.json.return_value = dict(total_count=3, limit=0, offset=0, **response)
        issues = self.redmine.issue.all()
        self.assertEqual(issues.limit, 0)
        self.assertEqual(issues.offset, 0)
        self.assertEqual(issues[0].id, 1)
        self.assertEqual(issues[1].id, 2)
        self.assertEqual(issues[2].id, 3)

    def test_offset_limit(self):
        self.response.json.return_value = {
            'total_count': 2, 'limit': 300, 'offset': 1, 'issues': response['issues'][1:3]}
        issues = self.redmine.issue.all()[1:300]
        self.assertEqual(issues.limit, 300)
        self.assertEqual(issues.offset, 1)
        self.assertEqual(issues[0].id, 2)
        self.assertEqual(issues[1].id, 3)
        self.assertEqual(issues[:1][0].id, 2)
        self.assertEqual(issues[1:][0].id, 3)
        self.assertEqual(issues[1:1][0].id, 3)

    def test_offset_limit_mimic(self):
        issues = self.redmine.issue.all()[1:3]
        self.assertEqual(issues.limit, 3)
        self.assertEqual(issues.offset, 1)
        self.assertEqual(issues[0].id, 2)
        self.assertEqual(issues[1].id, 3)

    def test_total_count(self):
        self.response.json.return_value = dict(total_count=3, limit=0, offset=0, **response)
        issues = self.redmine.issue.all()
        len(issues)
        self.assertEqual(issues.total_count, 3)

    def test_total_count_mimic(self):
        response_with_cf = {'issue': dict(custom_fields=[{'id': 1, 'value': 0}], **response['issues'][0])}
        self.response.json.return_value = response_with_cf
        issue = self.redmine.issue.get(1)
        self.assertEqual(issue.custom_fields.total_count, 1)

    def test_total_count_raise_exception_if_not_evaluated(self):
        self.assertRaises(exceptions.ResultSetTotalCountError, lambda: self.redmine.issue.all().total_count)

    def test_resultset_is_empty(self):
        self.response.json.return_value = {'limit': 100, 'issues': [], 'total_count': 0, 'offset': 0}
        issues = self.redmine.issue.all()
        self.assertEqual(len(issues), 0)
        self.assertEqual(list(issues), [])

    def test_sliced_resultset_is_empty(self):
        self.response.json.return_value = {'limit': 100, 'issues': [], 'total_count': 0, 'offset': 0}
        issues = self.redmine.issue.all()[:200]
        self.assertEqual(len(issues), 0)
        self.assertEqual(list(issues), [])

    def test_supports_iteration(self):
        issues = list(self.redmine.issue.all())
        self.assertEqual(issues[0].subject, 'Foo')
        self.assertEqual(issues[0].id, 1)
        self.assertEqual(issues[1].subject, 'Bar')
        self.assertEqual(issues[1].id, 2)

    def test_supports_len(self):
        self.assertEqual(len(self.redmine.issue.all()), 3)

    def test_get_method_resource_found(self):
        issues = self.redmine.issue.all().get(2)
        self.assertEqual(issues.id, 2)

    def test_get_method_resource_not_found(self):
        issues = self.redmine.issue.all().get(6)
        self.assertEqual(issues, None)

    def test_filter_method_nonexistant_attributes(self):
        issues = self.redmine.issue.all().filter(id=1, foo__exact=1)
        self.assertEqual(len(issues), 0)

    def test_filter_exact_lookup(self):
        issues = self.redmine.issue.all().filter(id=1, tracker_id__exact=1)
        self.assertEqual(issues[0].id, 1)
        self.assertEqual(len(issues), 1)

    def test_filter_in_lookup(self):
        issues = self.redmine.issue.all().filter(id__in=(1, 3))
        self.assertEqual(issues[0].id, 1)
        self.assertEqual(issues[1].id, 3)
        self.assertEqual(len(issues), 2)

    def test_update_method(self):
        issues = self.redmine.issue.all().update(subject='FooBar')
        self.assertEqual(issues[0].subject, 'FooBar')
        self.assertEqual(issues[1].subject, 'FooBar')
        self.assertEqual(issues[2].subject, 'FooBar')

    def test_delete_method(self):
        self.assertEqual(self.redmine.issue.all().delete(), True)

    def test_resourceset_is_picklable(self):
        import pickle
        issues = self.redmine.issue.all()
        unpickled_issues = pickle.loads(pickle.dumps(issues))
        self.assertEqual(issues[0]['subject'], unpickled_issues[0]['subject'])
        self.assertEqual(issues[1]['subject'], unpickled_issues[1]['subject'])
        self.assertEqual(issues[2]['subject'], unpickled_issues[2]['subject'])

    def test_values_method(self):
        issues = list(self.redmine.issue.all().values())
        self.assertEqual(issues[0]['subject'], 'Foo')
        self.assertEqual(issues[0]['id'], 1)
        self.assertEqual(issues[1]['subject'], 'Bar')
        self.assertEqual(issues[1]['id'], 2)
        self.assertEqual(issues[2]['subject'], 'Baz')
        self.assertEqual(issues[2]['id'], 3)

    def test_values_method_with_fields(self):
        issues = list(self.redmine.issue.all().values('subject', 'id'))
        self.assertEqual(len(issues[0]), 2)
        self.assertEqual(issues[0]['subject'], 'Foo')
        self.assertEqual(issues[0]['id'], 1)
        self.assertEqual(len(issues[1]), 2)
        self.assertEqual(issues[1]['subject'], 'Bar')
        self.assertEqual(issues[1]['id'], 2)
        self.assertEqual(len(issues[2]), 2)
        self.assertEqual(issues[2]['subject'], 'Baz')
        self.assertEqual(issues[2]['id'], 3)

    def test_values_list_method(self):
        issues = list(self.redmine.issue.all().values_list())
        self.assertIn('Foo', issues[0])
        self.assertIn(1, issues[0])
        self.assertIn('Bar', issues[1])
        self.assertIn(2, issues[1])
        self.assertIn('Baz', issues[2])
        self.assertIn(3, issues[2])

    def test_values_list_method_with_fields(self):
        issues = list(self.redmine.issue.all().values_list('id', 'subject'))
        self.assertEqual(len(issues[0]), 2)
        self.assertEqual(issues[0][0], 1)
        self.assertEqual(issues[0][1], 'Foo')
        self.assertEqual(len(issues[1]), 2)
        self.assertEqual(issues[1][0], 2)
        self.assertEqual(issues[1][1], 'Bar')
        self.assertEqual(len(issues[2]), 2)
        self.assertEqual(issues[2][0], 3)
        self.assertEqual(issues[2][1], 'Baz')

    def test_values_list_method_flattened(self):
        issues = list(self.redmine.issue.all().values_list('id', flat=True))
        self.assertEqual(issues[0], 1)
        self.assertEqual(issues[1], 2)
        self.assertEqual(issues[2], 3)

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_export(self):
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/'), '/foo/bar/issues.txt')

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_export_with_all_columns(self):
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns='all'), '/foo/bar/issues.txt')
        self.redmine.ver = (3, 3, 0)
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns='all'), '/foo/bar/issues.txt')

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_export_with_all_gui_columns(self):
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns='all_gui'), '/foo/bar/issues.txt')
        self.redmine.ver = (3, 3, 0)
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns='all_gui'), '/foo/bar/issues.txt')

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_export_with_all_gui_extra_columns(self):
        columns = ['all_gui']
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns=columns), '/foo/bar/issues.txt')
        self.redmine.ver = (3, 3, 0)
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns=columns), '/foo/bar/issues.txt')

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_export_with_custom_columns(self):
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.issue.all().export('txt', '/foo/bar/', columns=['status']), '/foo/bar/issues.txt')

    def test_export_not_supported_exception(self):
        self.assertRaises(exceptions.ExportNotSupported, lambda: self.redmine.custom_field.all().export('pdf'))

    def test_export_format_not_supported_exception(self):
        self.response.status_code = 406
        self.assertRaises(exceptions.ExportFormatNotSupportedError, lambda: self.redmine.issue.all().export('foo'))

    def test_export_reraises_unknown_error(self):
        self.response.status_code = 999
        self.assertRaises(exceptions.UnknownError, lambda: self.redmine.issue.all().export('foo'))

    def test_filter_no_filters_exception(self):
        self.assertRaises(exceptions.ResourceNoFiltersProvidedError, lambda: self.redmine.issue.all().filter())

    def test_filter_lookup_exception(self):
        self.assertRaises(exceptions.ResourceSetFilterLookupError, lambda: self.redmine.issue.all().filter(id__bar=1))

    def test_filter_bad_lookup_class_definition(self):
        from redminelib import lookups
        type('Foo', (lookups.Lookup,), {'lookup_name': 'foo'})
        self.assertRaises(NotImplementedError, lambda: self.redmine.issue.all().filter(id__foo=1))
        del lookups.registry['foo']

    def test_index_error_exception(self):
        self.assertRaises(exceptions.ResourceSetIndexError, lambda: self.redmine.issue.all()[6])



import warnings

from . import mock, BaseRedmineTestCase, Redmine

from redminelib import engines, resultsets, exceptions


class RedmineTestCase(BaseRedmineTestCase):
    def test_default_attributes(self):
        self.assertEqual(self.redmine.url, self.url)
        self.assertEqual(self.redmine.ver, None)
        self.assertEqual(self.redmine.timezone, None)
        self.assertEqual(self.redmine.date_format, '%Y-%m-%d')
        self.assertEqual(self.redmine.datetime_format, '%Y-%m-%dT%H:%M:%SZ')
        self.assertEqual(self.redmine.raise_attr_exception, True)
        self.assertEqual(self.redmine.engine.__class__, engines.DefaultEngine)

    def test_set_attributes_through_kwargs(self):
        from datetime import timezone
        FooEngine = type('FooEngine', (engines.BaseEngine,), {'create_session': lambda obj, **kwargs: None})
        redmine = Redmine(self.url, version='1.0.0', date_format='format', datetime_format='format',
                          timezone='+0000', raise_attr_exception=False, engine=FooEngine)
        self.assertEqual(redmine.url, self.url)
        self.assertEqual(redmine.ver, (1, 0, 0))
        self.assertEqual(redmine.timezone, timezone.utc)
        self.assertEqual(redmine.date_format, 'format')
        self.assertEqual(redmine.datetime_format, 'format')
        self.assertEqual(redmine.raise_attr_exception, False)
        self.assertEqual(redmine.engine.__class__, FooEngine)

    def test_timezone_exception(self):
        self.assertRaises(exceptions.TimezoneError, lambda: Redmine(self.url, timezone='foobar'))

    def test_engine_class_exception(self):
        self.assertRaises(exceptions.EngineClassError, lambda: Redmine(self.url, engine=type('Foo', (object,), {})))

    def test_session_impersonate(self):
        with self.redmine.session(impersonate='jsmith'):
            self.assertEqual(self.redmine.engine.requests['headers']['X-Redmine-Switch-User'], 'jsmith')
        self.assertRaises(KeyError, lambda: self.redmine.engine.requests['headers']['X-Redmine-Switch-User'])

    def test_session_key(self):
        with self.redmine.session(key='opa'):
            self.assertEqual(self.redmine.engine.requests['headers']['X-Redmine-API-Key'], 'opa')
        self.assertRaises(KeyError, lambda: self.redmine.engine.requests['headers']['X-Redmine-API-Key'])

    def test_session_username_password(self):
        with self.redmine.session(username='john', password='smith'):
            self.assertEqual(self.redmine.engine.requests['auth'], ('john', 'smith'))
        self.assertRaises(KeyError, lambda: self.redmine.engine.requests['auth'])

    def test_session_requests(self):
        self.redmine.engine.requests['cert'] = ('bar', 'baz')
        requests = {'verify': False, 'timeout': 2, 'cert': ('foo', 'bar'), 'params': {'foo': 'bar'}}
        with self.redmine.session(key='secret', requests=requests):
            self.assertEqual(self.redmine.engine.requests['headers'], {'X-Redmine-API-Key': 'secret'})
            self.assertEqual(self.redmine.engine.requests['params'], requests['params'])
            self.assertEqual(self.redmine.engine.requests['verify'], requests['verify'])
            self.assertEqual(self.redmine.engine.requests['timeout'], requests['timeout'])
            self.assertEqual(self.redmine.engine.requests['cert'], requests['cert'])
        self.assertEqual(self.redmine.engine.requests['params'], {})
        self.assertEqual(self.redmine.engine.requests['cert'], ('bar', 'baz'))
        self.assertRaises(KeyError, lambda: self.redmine.engine.requests['verify'])
        self.assertRaises(KeyError, lambda: self.redmine.engine.requests['timeout'])

    @mock.patch('os.path.isfile', mock.Mock())
    @mock.patch('os.path.getsize', mock.Mock())
    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_successful_file_upload(self):
        self.response.status_code = 201
        self.response.json.return_value = {'upload': {'id': 1, 'token': '123456'}}
        self.assertEqual(self.redmine.upload('foo', filename='foo.jpg')['token'], '123456')

    def test_successful_filestream_upload(self):
        from io import StringIO
        self.response.status_code = 201
        self.response.json.return_value = {'upload': {'id': 1, 'token': '456789'}}
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self.assertEqual(self.redmine.upload(StringIO(b'\xcf\x86oo'.decode('utf-8')))['token'], '456789')
            self.assertEqual(len(w), 1)
            self.assertIs(w[0].category, exceptions.PerformanceWarning)

    @mock.patch('redminelib.open', mock.mock_open(), create=True)
    def test_successful_file_download(self):
        self.response.status_code = 200
        self.response.iter_content = lambda chunk_size: (str(num) for num in range(0, 5))
        self.assertEqual(self.redmine.download(f'{self.url}/bar.txt', '/some/path/'), '/some/path/bar.txt')

    def test_successful_in_memory_file_download(self):
        self.response.status_code = 200
        self.response.iter_content = lambda: (str(num) for num in range(0, 5))
        self.assertEqual(''.join(self.redmine.download(f'{self.url}/bar.txt').iter_content()), '01234')

    def test_file_url_exception(self):
        self.response.status_code = 200
        self.assertRaises(exceptions.FileUrlError, lambda: self.redmine.download('https://bad_url', '/some/path'))

    def test_file_upload_no_file_exception(self):
        self.assertRaises(exceptions.NoFileError, lambda: self.redmine.upload('foo',))

    def test_file_upload_file_object_exception(self):
        f = type('FileObject', (), {'close': lambda obj: None})()
        self.assertRaises(exceptions.FileObjectError, lambda: self.redmine.upload(f))

    def test_file_upload_not_supported_exception(self):
        self.redmine.ver = (1, 0, 0)
        self.assertRaises(exceptions.VersionMismatchError, lambda: self.redmine.upload('foo',))

    def test_auth(self):
        self.redmine.username = 'john'
        self.redmine.password = 'qwerty'
        self.response.status_code = 200
        self.response.json.return_value = {'user': {'firstname': 'John', 'lastname': 'Smith', 'id': 1}}
        self.assertEqual(self.redmine.auth().firstname, 'John')

    def test_search(self):
        self.response.json.return_value = {'total_count': 6, 'offset': 0, 'limit': 0, 'results': [
            {'id': 1, 'title': 'Foo', 'type': 'issue'},
            {'id': 2, 'title': 'Bar', 'type': 'issue closed'},
            {'id': 3, 'title': 'Foo', 'type': 'project'},
            {'id': 4, 'title': 'Foo', 'type': 'news'},
            {'id': 5, 'title': 'Foo', 'type': 'wiki-page'},
            {'id': 6, 'title': 'Foo', 'type': 'document'},
        ]}
        results = self.redmine.search('foo')
        self.assertIsInstance(results['issues'], resultsets.ResourceSet)
        self.assertEqual(len(results['issues']), 2)
        self.assertIsInstance(results['projects'], resultsets.ResourceSet)
        self.assertEqual(len(results['projects']), 1)
        self.assertIsInstance(results['news'], resultsets.ResourceSet)
        self.assertEqual(len(results['news']), 1)
        self.assertIsInstance(results['wiki_pages'], resultsets.ResourceSet)
        self.assertEqual(len(results['wiki_pages']), 1)
        self.assertIsInstance(results['unknown'], dict)
        self.assertEqual(len(results['unknown']['document']), 1)

    def test_search_without_unknown(self):
        self.response.json.return_value = {'total_count': 1, 'offset': 0, 'limit': 0, 'results': [
            {'id': 1, 'title': 'Foo', 'type': 'issue'}]}
        results = self.redmine.search('foo')
        self.assertIsInstance(results['issues'], resultsets.ResourceSet)
        self.assertEqual(len(results['issues']), 1)

    def test_unsupported_version_format_exception(self):
        self.assertRaises(exceptions.VersionFormatError, lambda: Redmine(self.url, version='4.1'))

    def test_search_not_supported_exception(self):
        self.redmine.ver = (1, 0, 0)
        self.assertRaises(exceptions.VersionMismatchError, lambda: self.redmine.search('foo'))

    def test_redmine_is_picklable(self):
        import pickle
        redmine = pickle.loads(pickle.dumps(self.redmine))
        self.assertEqual(redmine.url, self.redmine.url)
        self.assertEqual(redmine.ver, self.redmine.ver)
        self.assertEqual(redmine.date_format, self.redmine.date_format)
        self.assertEqual(redmine.datetime_format, self.redmine.datetime_format)
        self.assertEqual(redmine.raise_attr_exception, self.redmine.raise_attr_exception)
